from data_gateway_sdk.protocols.udp_wrapper.unbp_len import UnbpUplinkLenData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_unbp_uplink_len() -> None:
    case_serialized = bytes.fromhex("3bd22200000000037f3f160b973d76")
    assert UnbpUplinkLenData(mac=34, message_id=0, payload=[127, 63, 22], crc=1983747851) == UnbpUplinkLenData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == UnbpUplinkLenData.serialize(UnbpUplinkLenData(mac=34, message_id=0, payload=[127, 63, 22], crc=1983747851))  # noqa: E501
    case_serialized = bytes.fromhex("3bd20000000000002e82d981")  # overflow
    assert UnbpUplinkLenData(mac=0, message_id=0, payload=[], crc=2178515502) == UnbpUplinkLenData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == UnbpUplinkLenData.serialize(UnbpUplinkLenData(crc=2178515502, mac=0, message_id=0, payload=[]))  # noqa: E501
    case_serialized = bytes.fromhex("3bd222000000000868e4eab106a33a78a6258953")  # overflow
    assert UnbpUplinkLenData(mac=34, message_id=0, payload=[104, 228, 234, 177, 6, 163, 58, 120], crc=1401496998) == UnbpUplinkLenData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == UnbpUplinkLenData.serialize(UnbpUplinkLenData(crc=1401496998, mac=34, message_id=0, payload=[104, 228, 234, 177, 6, 163, 58, 120]))  # noqa: E501


def test_unbp_uplink_len_multiple() -> None:
    case_serialized = bytes.fromhex("3bd298248000001068e4eab106a33a7868e4eab106a33a78ebad25df")  # overflow
    assert UnbpUplinkLenData(mac=8397976, message_id=0, payload=[104, 228, 234, 177, 6, 163, 58, 120, 104, 228, 234, 177, 6, 163, 58, 120], crc=3743788523) == UnbpUplinkLenData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == UnbpUplinkLenData.serialize(UnbpUplinkLenData(crc=3743788523, mac=8397976, message_id=0, payload=[104, 228, 234, 177, 6, 163, 58, 120, 104, 228, 234, 177, 6, 163, 58, 120]))  # noqa: E501



from decimal import Decimal
from datetime import timedelta, datetime, time, tzinfo
from typing import List, Any, Tuple, Dict

from data_aggregator_sdk.constants.enums import JournalDataType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, ResourceType, \
    IntegrationV0MessageGeneration, CounterType

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_tariff_generated
#
# RESULT int:        1636017705544601932779225793
# RESULT bin:  MSB   00000000000000000000000000000000000001010100100101001000001110011000011101100000000111100000111100110010010100000000001011000001   LSB                             # noqa: E501
# RESULT hex:  LE    c1 02 50 32 0f 1e 60 87 39 48 49 05 00 00 00 00
#
#
# name                  type       size  value(int)                                                                                                                        data(bits)   # noqa: E501
# --------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL  u7            7          65                                                                                                                           1000001   # noqa: E501
# packet_type_id.0.DFF  bool          1           1                                                                                                                          1          # noqa: E501
# packet_type_id.1.VAL  u2            2           2                                                                                                                        10           # noqa: E501
# packet_type_id.1.DFF  bool          1           0                                                                                                                       0             # noqa: E501
# energy_is_reactive    bool          1           0                                                                                                                      0              # noqa: E501
# days_ago              timedelta     7           0                                                                                                               0000000               # noqa: E501
# valid                 bool          1           0                                                                                                              0                      # noqa: E501
# tariff_mask.0.tariff  bool          1           1                                                                                                             1                       # noqa: E501
# tariff_mask.1.tariff  bool          1           0                                                                                                            0                        # noqa: E501
# tariff_mask.2.tariff  bool          1           1                                                                                                           1                         # noqa: E501
# tariff_mask.3.tariff  bool          1           0                                                                                                          0                          # noqa: E501
# tariff_mask.4.tariff  bool          1           0                                                                                                         0                           # noqa: E501
# tariff_mask.5.tariff  bool          1           1                                                                                                        1                            # noqa: E501
# tariff_mask.6.tariff  bool          1           0                                                                                                       0                             # noqa: E501
# tariff_mask.7.tariff  bool          1           0                                                                                                      0                              # noqa: E501
# slot_0                u25          25      123123                                                                             0000000011110000011110011                               # noqa: E501
# slot_1                u25          25     4312123                                                    0010000011100110000111011                                                        # noqa: E501
# slot_2                u25          25        5413                           0000000000001010100100101
# slot_3                u25          25           0  0000000000000000000000000


class SmpmUlDeviceEnergy16BTariffGeneratedData(Packet):
    energy_is_reactive: bool
    days_ago: timedelta
    valid: bool
    tariff_mask: Tuple[bool, bool, bool, bool, bool, bool, bool, bool]
    slot_0: int
    slot_1: int
    slot_2: int
    slot_3: int

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((65) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((2) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.energy_is_reactive, bool)
        result |= ((int(data.energy_is_reactive)) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 127
        result |= ((days_ago_tmp1) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.valid, bool)
        result |= ((int(data.valid)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.tariff_mask, tuple) and len(data.tariff_mask) == 8
        assert isinstance(data.tariff_mask[0], bool)
        result |= ((int(data.tariff_mask[0])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.tariff_mask[1], bool)
        result |= ((int(data.tariff_mask[1])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.tariff_mask[2], bool)
        result |= ((int(data.tariff_mask[2])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.tariff_mask[3], bool)
        result |= ((int(data.tariff_mask[3])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.tariff_mask[4], bool)
        result |= ((int(data.tariff_mask[4])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.tariff_mask[5], bool)
        result |= ((int(data.tariff_mask[5])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.tariff_mask[6], bool)
        result |= ((int(data.tariff_mask[6])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.tariff_mask[7], bool)
        result |= ((int(data.tariff_mask[7])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.slot_0, int)
        result |= (((data.slot_0) & 33554431) & (2 ** (25) - 1)) << size
        size += 25
        assert isinstance(data.slot_1, int)
        result |= (((data.slot_1) & 33554431) & (2 ** (25) - 1)) << size
        size += 25
        assert isinstance(data.slot_2, int)
        result |= (((data.slot_2) & 33554431) & (2 ** (25) - 1)) << size
        size += 25
        assert isinstance(data.slot_3, int)
        result |= (((data.slot_3) & 33554431) & (2 ** (25) - 1)) << size
        size += 25
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BTariffGeneratedData':
        result__el_tmp2: Dict[str, Any] = dict()
        if 65 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 2 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp2["energy_is_reactive"] = bool(buf.shift(1))
        result__el_tmp2["days_ago"] = timedelta(seconds=buf.shift(7) * 86400)
        result__el_tmp2["valid"] = bool(buf.shift(1))
        tariff_mask_tmp3: List[bool] = []
        tariff_mask__item_tmp4 = bool(buf.shift(1))
        tariff_mask_tmp3.append(tariff_mask__item_tmp4)
        tariff_mask__item_tmp4 = bool(buf.shift(1))
        tariff_mask_tmp3.append(tariff_mask__item_tmp4)
        tariff_mask__item_tmp4 = bool(buf.shift(1))
        tariff_mask_tmp3.append(tariff_mask__item_tmp4)
        tariff_mask__item_tmp4 = bool(buf.shift(1))
        tariff_mask_tmp3.append(tariff_mask__item_tmp4)
        tariff_mask__item_tmp4 = bool(buf.shift(1))
        tariff_mask_tmp3.append(tariff_mask__item_tmp4)
        tariff_mask__item_tmp4 = bool(buf.shift(1))
        tariff_mask_tmp3.append(tariff_mask__item_tmp4)
        tariff_mask__item_tmp4 = bool(buf.shift(1))
        tariff_mask_tmp3.append(tariff_mask__item_tmp4)
        tariff_mask__item_tmp4 = bool(buf.shift(1))
        tariff_mask_tmp3.append(tariff_mask__item_tmp4)
        result__el_tmp2["tariff_mask"] = tuple(tariff_mask_tmp3)
        result__el_tmp2["slot_0"] = buf.shift(25) + 0
        result__el_tmp2["slot_1"] = buf.shift(25) + 0
        result__el_tmp2["slot_2"] = buf.shift(25) + 0
        result__el_tmp2["slot_3"] = buf.shift(25) + 0
        result = SmpmUlDeviceEnergy16BTariffGeneratedData(**result__el_tmp2)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        slots = [self.slot_0, self.slot_1, self.slot_2, self.slot_3]
        generations = []
        for tariff_number, tariff in enumerate(self.tariff_mask, 1):
            if tariff and slots:
                generations.append(
                    IntegrationV0MessageGeneration(
                        tariff=tariff_number,
                        value=Decimal(str(slots.pop(0))),
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.ENERGY_REACTIVE if self.energy_is_reactive else ResourceType.ENERGY_ACTIVE,
                        overloading_value=None,
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                )
        return [
            IntegrationV0MessageData(
                is_valid=self.valid,
                dt=days_ago_calculation(received_at, device_tz, time(0), self.days_ago),
                generation=generations,
            ),
        ]



from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_12b_counter_volume import \
    SmpmUlDeviceJupiter12BCounterVolumeData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_jupiter_12b_counter_volume() -> None:
    case_serialized = bytes.fromhex("d5f9fffffffbffffff43d61d00000000")
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=2147483.647, volume_channel_2=2147483.647, battery_volts=2.0, temperature=23, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=2147483.647, volume_channel_2=2147483.647, battery_volts=2.0, temperature=23, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True))  # noqa: E501
    case_serialized = bytes.fromhex("d5010000000000000000f81f00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=0.0, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=92, volume_channel_1=0.0, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("d5010000000000000000f81f00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=0.0, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=92, volume_channel_1=0.0, volume_channel_2=8589934.592))  # noqa: E501
    case_serialized = bytes.fromhex("d5010000000000000000f81f00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=0.0, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=92, volume_channel_1=8589934.592, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("d5010000000000000000f81f00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=0.0, temperature=92, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=0.0, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=92, volume_channel_1=8589934.592, volume_channel_2=8589934.592))  # noqa: E501
    case_serialized = bytes.fromhex("d5010000000000000000001c00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=0.0, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=-35, volume_channel_1=0.0, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("d5010000000000000000001c00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=0.0, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=-35, volume_channel_1=0.0, volume_channel_2=8589934.592))  # noqa: E501
    case_serialized = bytes.fromhex("d5010000000000000000001c00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=0.0, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=-35, volume_channel_1=8589934.592, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("d5010000000000000000001c00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=0.0, temperature=-35, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=0.0, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=-35, volume_channel_1=8589934.592, volume_channel_2=8589934.592))  # noqa: E501
    case_serialized = bytes.fromhex("d50100000000000000f0ff1f00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=2.54, temperature=92, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=2.54, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=92, volume_channel_1=0.0, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("d50100000000000000f0ff1f00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=2.54, temperature=92, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=2.54, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=92, volume_channel_1=0.0, volume_channel_2=8589934.592))  # noqa: E501
    case_serialized = bytes.fromhex("d50100000000000000f0ff1f00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=2.54, temperature=92, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=2.54, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=92, volume_channel_1=8589934.592, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("d50100000000000000f0ff1f00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=2.54, temperature=92, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=2.54, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=92, volume_channel_1=8589934.592, volume_channel_2=8589934.592))  # noqa: E501
    case_serialized = bytes.fromhex("d50100000000000000f0071c00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=2.54, temperature=-35, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=2.54, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=-35, volume_channel_1=0.0, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("d50100000000000000f0071c00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=2.54, temperature=-35, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=2.54, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=-35, volume_channel_1=0.0, volume_channel_2=8589934.592))  # noqa: E501
    case_serialized = bytes.fromhex("d50100000000000000f0071c00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=2.54, temperature=-35, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=2.54, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=-35, volume_channel_1=8589934.592, volume_channel_2=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("d50100000000000000f0071c00000000")  # overflow
    assert SmpmUlDeviceJupiter12BCounterVolumeData(volume_channel_1=0.0, volume_channel_2=0.0, battery_volts=2.54, temperature=-35, event_reset=True, event_low_battery_level=True, event_low_ambient_temperature=True) == SmpmUlDeviceJupiter12BCounterVolumeData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter12BCounterVolumeData.serialize(SmpmUlDeviceJupiter12BCounterVolumeData(battery_volts=2.54, event_low_ambient_temperature=True, event_low_battery_level=True, event_reset=True, temperature=-35, volume_channel_1=8589934.592, volume_channel_2=8589934.592))  # noqa: E501



from data_gateway_sdk.protocols.smpm.smpm_ul_device_dl_answer import SmpmUlDeviceDlAnswerData, \
    SmpmUlDeviceDlAnswerDataDownlinkPacketId
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_dl_answer() -> None:
    case_serialized = bytes.fromhex("030100ffffff7f07")
    assert SmpmUlDeviceDlAnswerData(downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.GET_ECHO, downlink_packet_crc=2147483647, answer_packets_count=7) == SmpmUlDeviceDlAnswerData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceDlAnswerData.serialize(SmpmUlDeviceDlAnswerData(downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.GET_ECHO, downlink_packet_crc=2147483647, answer_packets_count=7))  # noqa: E501
    case_serialized = bytes.fromhex("03aa000000000000")  # overflow
    assert SmpmUlDeviceDlAnswerData(downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.SET_RELAY, downlink_packet_crc=0, answer_packets_count=0) == SmpmUlDeviceDlAnswerData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceDlAnswerData.serialize(SmpmUlDeviceDlAnswerData(answer_packets_count=0, downlink_packet_crc=0, downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.SET_RELAY))  # noqa: E501
    case_serialized = bytes.fromhex("03aa00ffffffff00")  # overflow
    assert SmpmUlDeviceDlAnswerData(downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.SET_RELAY, downlink_packet_crc=4294967295, answer_packets_count=0) == SmpmUlDeviceDlAnswerData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceDlAnswerData.serialize(SmpmUlDeviceDlAnswerData(answer_packets_count=0, downlink_packet_crc=4294967295, downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.SET_RELAY))  # noqa: E501
    case_serialized = bytes.fromhex("03aa00000000000f")  # overflow
    assert SmpmUlDeviceDlAnswerData(downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.SET_RELAY, downlink_packet_crc=0, answer_packets_count=15) == SmpmUlDeviceDlAnswerData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceDlAnswerData.serialize(SmpmUlDeviceDlAnswerData(answer_packets_count=15, downlink_packet_crc=0, downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.SET_RELAY))  # noqa: E501
    case_serialized = bytes.fromhex("03aa00ffffffff0f")  # overflow
    assert SmpmUlDeviceDlAnswerData(downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.SET_RELAY, downlink_packet_crc=4294967295, answer_packets_count=15) == SmpmUlDeviceDlAnswerData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceDlAnswerData.serialize(SmpmUlDeviceDlAnswerData(answer_packets_count=15, downlink_packet_crc=4294967295, downlink_packet_id=SmpmUlDeviceDlAnswerDataDownlinkPacketId.SET_RELAY))  # noqa: E501



from data_gateway_sdk.protocols.water5_v_nero.water5_device_magnet import Water5DeviceMagnetData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_water5_device_magnet() -> None:
    case = Water5DeviceMagnetData()  # noqa: E501
    assert case == Water5DeviceMagnetData.parse(BufRef(case.serialize()))
    case = Water5DeviceMagnetData()  # noqa: E501
    assert case == Water5DeviceMagnetData.parse(BufRef(case.serialize()))



from ul_unipipeline.message.uni_message import UniMessage

from data_gateway_sdk.protocols.nero_bs_packet.http_nero_bs_packet import HttpV0NeroBsPacket


class InputBsHttpNeroV0Message(UniMessage):
    data: HttpV0NeroBsPacket



import random
import typing
from enum import IntEnum, unique

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   smpm_dl_device_energy_8b_set_relay
#
# RESULT int:        426
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00000000 00000000 00000001 10101010   LSB
# RESULT hex:  LE    aa 01 00 00 00 00 00 00
#
# name                  type                            size  value(int)                                                        data(bits)
# --------------------  ------------------------------  ----  ----------  ----------------------------------------------------------------
# packet_type_id.0.VAL  u7                                 7          42                                                           0101010
# packet_type_id.0.DFF  bool                               1           1                                                          1
# packet_type_id.1.VAL  u2                                 2           1                                                        01
# packet_type_id.1.DFF  bool                               1           0                                                       0
# state                 bool                               1           0                                                      0
# relay_id              SmpmDlDeviceEnergy8bSetRelayId     8           0                                              00000000
# RESERVED              u44                               44           0  00000000000000000000000000000000000000000000
# Reserved space is filled with random bits in DOWNLINK packets !


@unique
class SmpmDlDeviceEnergy8bSetRelayId(IntEnum):
    ALL = 0
    RELAY_ID_1 = 1
    RELAY_ID_2 = 2
    RELAY_ID_3 = 3
    RELAY_ID_4 = 4
    RELAY_ID_5 = 5
    RELAY_ID_6 = 6
    RELAY_ID_7 = 7


class SmpmDlDeviceEnergy8BSetRelayData(Packet):
    state: bool
    relay_id: SmpmDlDeviceEnergy8bSetRelayId

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((42) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.state, bool)
        result |= ((int(data.state)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.relay_id, SmpmDlDeviceEnergy8bSetRelayId)
        result |= ((data.relay_id.value) & (2 ** (8) - 1)) << size
        size += 8
        result |= ((random.getrandbits(44)) & (2 ** 44 - 1)) << size
        size += 44
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmDlDeviceEnergy8BSetRelayData':
        result__el_tmp1: typing.Dict[str, typing.Any] = dict()
        if 42 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["state"] = bool(buf.shift(1))
        result__el_tmp1["relay_id"] = SmpmDlDeviceEnergy8bSetRelayId(buf.shift(8))
        result = SmpmDlDeviceEnergy8BSetRelayData(**result__el_tmp1)
        buf.shift(44)
        return result



from enum import Enum, unique
from typing import Any, List

from data_gateway_sdk.errors import DataGatewayDeviceProtocolParsingError
from data_gateway_sdk.protocols.device_packet.device_packet import DevicePacket
from data_gateway_sdk.protocols.device_packet.ncp_smp_device_packet import NcpSmpV0DevicePacket
from data_gateway_sdk.protocols.device_packet.smp_device_packet import SmpV0DevicePacket
from data_gateway_sdk.protocols.device_packet.smpm_device_packet import SmpMGasMeterV0DevicePacket, \
    SmpMEnergyMeterV0DevicePacket, SmpMJupiter08BV0DevicePacket, SmpMJupiter12BV0DevicePacket, \
    SmpMWaterMeter08BV0DevicePacket, SmpMWaterMeter12BV0DevicePacket, SmpMWaterMeter04BV0DevicePacket, \
    SmpMJupiter16BV0DevicePacket, SmpMWaterMeter16BV0DevicePacket, \
    SmpMInternalInfoDataV0DevicePacket, SmpMHeatProxyMeter16BV0DevicePacket, SmpMHeatGroupMeterV0DevicePacket, \
    SmpMEnergyMeterV1DevicePacket, SmpMDownlinkV0DevicePacket
from data_gateway_sdk.protocols.device_packet.water5_device_packet import Water5NeroV0DevicePacket, \
    Water5JupiterFreeScaleV0DevicePacket, Water5JupiterSTMV0DevicePacket, Water5FluoSTMV0DevicePacket, \
    Water5FluoFreeScaleV0DevicePacket, Water5FluoAV0DevicePacket, Water5GasV0DevicePacket, \
    Water5JupiterLoraV0DevicePacket, Water5FluoSV0DevicePacket


@unique
class DeviceProtocolApiType(Enum):
    ARVAS_API_V0 = 'ARVAS_API_V0'
    UNIVERSAL_API_V0 = 'UNIVERSAL_API_V0'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


DEVICE_PROTOCOL_API_TYPE_LIST = [device_protocol_api_type.value for device_protocol_api_type in DeviceProtocolApiType]


@unique
class DeviceProtocolType(Enum):  # DO NOT CHANGE VALUE because it should be compatible with data_aggregator_sdk db !!!
    WATER5_V_NERO_V0 = 'WATER5_V_NERO_V0'
    NCP_SMP_V0 = 'NCP_SMP_V0'  # only for jupiter NbIoT
    SMP_V0 = 'SMP_V0'  # only for jupiter NbIoT
    SMP_M_GAS_METER_V0 = 'SMP_M_GAS_METER_V0'
    SMP_M_ENERGY_METER_V0 = 'SMP_M_ENERGY_METER_V0'
    SMP_M_ENERGY_METER_V1 = 'SMP_M_ENERGY_METER_V1'
    SMP_M_JUPITER_08B_V0 = 'SMP_M_JUPITER_08B_V0'
    SMP_M_JUPITER_12B_V0 = 'SMP_M_JUPITER_12B_V0'
    SMP_M_JUPITER_16B_V0 = 'SMP_M_JUPITER_16B_V0'
    SMP_M_WATER_METER_04B_V0 = 'SMP_M_WATER_METER_04B_V0'
    SMP_M_WATER_METER_08B_V0 = 'SMP_M_WATER_METER_08B_V0'
    SMP_M_WATER_METER_12B_V0 = 'SMP_M_WATER_METER_12B_V0'
    SMP_M_WATER_METER_16B_V0 = 'SMP_M_WATER_METER_16B_V0'
    SMP_M_HEAT_PROXY_METER_16B_V0 = 'SMP_M_HEAT_PROXY_METER_16B_V0'
    SMP_M_HEAT_GROUP_METER_V0 = 'SMP_M_HEAT_GROUP_METER_V0'
    WATER5_V_JUPITER_FREESCALE_V0 = 'WATER5_V_JUPITER_FREESCALE_V0'
    WATER5_V_JUPITER_STM_V0 = 'WATER5_V_JUPITER_STM_V0'
    WATER5_V_FLUO_STM_V0 = 'WATER5_V_FLUO_STM_V0'
    WATER5_V_FLUO_FREESCALE_V0 = 'WATER5_V_FLUO_FREESCALE_V0'
    WATER5_V_FLUO_A_V0 = 'WATER5_V_FLUO_A_V0'
    WATER5_V_FLUO_S_V0 = 'WATER5_V_FLUO_S_V0'
    WATER5_V_GAS_V0 = 'WATER5_V_GAS_V0'
    WATER5_V_JUPITER_LORA_V0 = 'WATER5_V_JUPITER_LORA_V0'
    WATER5_V_FLUO_LORA_V0 = 'WATER5_V_FLUO_LORA_V0'
    SMP_M_INTERNAL_INFO_DATA = 'SMP_M_INTERNAL_INFO_DATA'
    SMP_M_DOWNLINK_V0 = 'SMP_M_DOWNLINK_V0'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'

    def parse(self, payload: bytes, **kwargs: Any) -> List[DevicePacket]:
        assert isinstance(payload, bytes), f'payload must be type of {bytes}, got type {type(payload)}'
        try:
            return DEVICE_PROTOCOL_MAP[self].parse(payload, **kwargs)  # type: ignore
        except DataGatewayDeviceProtocolParsingError:
            raise
        except Exception as e:  # noqa: B902
            raise DataGatewayDeviceProtocolParsingError(f'{self.value}: invalid payload', e)


DEVICE_PROTOCOL_MAP = {
    DeviceProtocolType.WATER5_V_NERO_V0: Water5NeroV0DevicePacket,
    DeviceProtocolType.NCP_SMP_V0: NcpSmpV0DevicePacket,
    DeviceProtocolType.SMP_V0: SmpV0DevicePacket,
    DeviceProtocolType.SMP_M_GAS_METER_V0: SmpMGasMeterV0DevicePacket,
    DeviceProtocolType.SMP_M_ENERGY_METER_V0: SmpMEnergyMeterV0DevicePacket,
    DeviceProtocolType.SMP_M_ENERGY_METER_V1: SmpMEnergyMeterV1DevicePacket,
    DeviceProtocolType.SMP_M_JUPITER_08B_V0: SmpMJupiter08BV0DevicePacket,
    DeviceProtocolType.SMP_M_JUPITER_12B_V0: SmpMJupiter12BV0DevicePacket,
    DeviceProtocolType.SMP_M_JUPITER_16B_V0: SmpMJupiter16BV0DevicePacket,
    DeviceProtocolType.SMP_M_WATER_METER_04B_V0: SmpMWaterMeter04BV0DevicePacket,
    DeviceProtocolType.SMP_M_WATER_METER_08B_V0: SmpMWaterMeter08BV0DevicePacket,
    DeviceProtocolType.SMP_M_WATER_METER_12B_V0: SmpMWaterMeter12BV0DevicePacket,
    DeviceProtocolType.SMP_M_WATER_METER_16B_V0: SmpMWaterMeter16BV0DevicePacket,

    DeviceProtocolType.SMP_M_HEAT_PROXY_METER_16B_V0: SmpMHeatProxyMeter16BV0DevicePacket,
    DeviceProtocolType.SMP_M_HEAT_GROUP_METER_V0: SmpMHeatGroupMeterV0DevicePacket,

    DeviceProtocolType.WATER5_V_JUPITER_FREESCALE_V0: Water5JupiterFreeScaleV0DevicePacket,
    DeviceProtocolType.WATER5_V_JUPITER_STM_V0: Water5JupiterSTMV0DevicePacket,
    DeviceProtocolType.WATER5_V_FLUO_STM_V0: Water5FluoSTMV0DevicePacket,
    DeviceProtocolType.WATER5_V_FLUO_FREESCALE_V0: Water5FluoFreeScaleV0DevicePacket,
    DeviceProtocolType.WATER5_V_FLUO_A_V0: Water5FluoAV0DevicePacket,
    DeviceProtocolType.WATER5_V_FLUO_S_V0: Water5FluoSV0DevicePacket,
    DeviceProtocolType.WATER5_V_GAS_V0: Water5GasV0DevicePacket,

    DeviceProtocolType.WATER5_V_JUPITER_LORA_V0: Water5JupiterLoraV0DevicePacket,
    DeviceProtocolType.WATER5_V_FLUO_LORA_V0: Water5JupiterLoraV0DevicePacket,

    DeviceProtocolType.SMP_M_INTERNAL_INFO_DATA: SmpMInternalInfoDataV0DevicePacket,

    DeviceProtocolType.SMP_M_DOWNLINK_V0: SmpMDownlinkV0DevicePacket,
}

assert set(DEVICE_PROTOCOL_MAP.keys()) == set(DeviceProtocolType)



from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch1 import Water5DeviceWeeklyCh1Data
from data_gateway_sdk.utils.buf_ref import BufRef


def test_water5_device_weekly_ch1() -> None:
    case = Water5DeviceWeeklyCh1Data(value=12313.0, battery_fract=67, battery_int=3)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=0, battery_int=2, value=0.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=0, battery_int=2, value=134217.727)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=0, battery_int=2, value=12313.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=0, battery_int=3, value=0.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=0, battery_int=3, value=134217.727)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=0, battery_int=3, value=12313.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=67, battery_int=2, value=0.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=67, battery_int=2, value=134217.727)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=67, battery_int=2, value=12313.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=67, battery_int=3, value=0.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=67, battery_int=3, value=134217.727)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=67, battery_int=3, value=12313.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=127, battery_int=2, value=0.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=127, battery_int=2, value=134217.727)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=127, battery_int=2, value=12313.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=127, battery_int=3, value=0.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=127, battery_int=3, value=134217.727)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceWeeklyCh1Data(battery_fract=127, battery_int=3, value=12313.0)  # noqa: E501
    assert case == Water5DeviceWeeklyCh1Data.parse(BufRef(case.serialize()))



from decimal import Decimal
from datetime import datetime, tzinfo
from typing import Dict, Any, List

from data_aggregator_sdk.constants.enums import SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_electricity_params
#
# RESULT int:        47268593044936706359241417748435900
# RESULT bin:  MSB   00000000 00001001 00011010 10000101 10000101 10000101 01101110 00110111 00011011 10000100 10001101 00010010 00110100 01001000 11010011 10111100   LSB
# RESULT hex:  LE    bc d3 48 34 12 8d 84 1b 37 6e 85 85 85 1a 09 00
#
# name                  type    size  value(int)                                                                                                                        data(bits)
# --------------------  ------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL  u7         7          60                                                                                                                           0111100
# packet_type_id.0.DFF  bool       1           1                                                                                                                          1
# packet_type_id.1.VAL  u2         2           3                                                                                                                        11
# packet_type_id.1.DFF  bool       1           0                                                                                                                       0
# current_ch_1          uf14p2    14        2330                                                                                                         00100100011010
# current_ch_2          uf14p2    14        2330                                                                                           00100100011010
# current_ch_3          uf14p2    14        2330                                                                             00100100011010
# voltage_ch1           u9         9         220                                                                    011011100
# voltage_ch2           u9         9         220                                                           011011100
# voltage_ch3           u9         9         220                                                  011011100
# k_ch1                 uf8p2      8         133                                          10000101
# k_ch2                 uf8p2      8         133                                  10000101
# k_ch3                 uf8p2      8         133                          10000101
# freq                  uf12p2    12        2330              100100011010
# RESERVED              u12       12           0  000000000000


class SmpmUlDeviceEnergy16BElectricityParamsData(Packet):
    current_ch_1: float
    current_ch_2: float
    current_ch_3: float
    voltage_ch1: int
    voltage_ch2: int
    voltage_ch3: int
    k_ch1: float
    k_ch2: float
    k_ch3: float
    freq: float

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((60) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((3) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.current_ch_1, (int, float))
        assert 0.0 <= data.current_ch_1 <= 163.83
        result |= ((int(round(float(data.current_ch_1) * 100.0, 0))) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.current_ch_2, (int, float))
        assert 0.0 <= data.current_ch_2 <= 163.83
        result |= ((int(round(float(data.current_ch_2) * 100.0, 0))) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.current_ch_3, (int, float))
        assert 0.0 <= data.current_ch_3 <= 163.83
        result |= ((int(round(float(data.current_ch_3) * 100.0, 0))) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.voltage_ch1, int)
        assert 0 <= data.voltage_ch1 <= 511
        result |= ((data.voltage_ch1) & (2 ** (9) - 1)) << size
        size += 9
        assert isinstance(data.voltage_ch2, int)
        assert 0 <= data.voltage_ch2 <= 511
        result |= ((data.voltage_ch2) & (2 ** (9) - 1)) << size
        size += 9
        assert isinstance(data.voltage_ch3, int)
        assert 0 <= data.voltage_ch3 <= 511
        result |= ((data.voltage_ch3) & (2 ** (9) - 1)) << size
        size += 9
        assert isinstance(data.k_ch1, (int, float))
        assert -1.0 <= data.k_ch1 <= 1.5499999999999998
        result |= ((int(round(float(data.k_ch1 - -1.0) * 100.0, 0))) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.k_ch2, (int, float))
        assert -1.0 <= data.k_ch2 <= 1.5499999999999998
        result |= ((int(round(float(data.k_ch2 - -1.0) * 100.0, 0))) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.k_ch3, (int, float))
        assert -1.0 <= data.k_ch3 <= 1.5499999999999998
        result |= ((int(round(float(data.k_ch3 - -1.0) * 100.0, 0))) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.freq, (int, float))
        assert 30.0 <= data.freq <= 70.95
        result |= ((int(round(float(data.freq - 30.0) * 100.0, 0))) & (2 ** (12) - 1)) << size
        size += 12
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BElectricityParamsData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 60 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 3 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["current_ch_1"] = round(buf.shift(14) / 100.0, 2)
        result__el_tmp1["current_ch_2"] = round(buf.shift(14) / 100.0, 2)
        result__el_tmp1["current_ch_3"] = round(buf.shift(14) / 100.0, 2)
        result__el_tmp1["voltage_ch1"] = buf.shift(9) + 0
        result__el_tmp1["voltage_ch2"] = buf.shift(9) + 0
        result__el_tmp1["voltage_ch3"] = buf.shift(9) + 0
        result__el_tmp1["k_ch1"] = round(buf.shift(8) / 100.0 + -1.0, 2)
        result__el_tmp1["k_ch2"] = round(buf.shift(8) / 100.0 + -1.0, 2)
        result__el_tmp1["k_ch3"] = round(buf.shift(8) / 100.0 + -1.0, 2)
        result__el_tmp1["freq"] = round(buf.shift(12) / 100.0 + 30.0, 2)
        result = SmpmUlDeviceEnergy16BElectricityParamsData(**result__el_tmp1)
        buf.shift(12)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        sensors = [IntegrationV0MessageSensor(
            channel_id=1,
            sensor_type=SensorType.ENERGY_METER_CURRENT,
            value=Decimal(str(self.current_ch_1)),
        )]

        if self.current_ch_2 and self.current_ch_3:
            sensors.extend(
                [
                    IntegrationV0MessageSensor(
                        channel_id=1,
                        sensor_type=SensorType.ENERGY_METER_CURRENT,
                        value=Decimal(str(self.current_ch_2)),
                    ),
                    IntegrationV0MessageSensor(
                        channel_id=1,
                        sensor_type=SensorType.ENERGY_METER_CURRENT,
                        value=Decimal(str(self.current_ch_3)),
                    ),
                ],
            )

        sensors.append(
            IntegrationV0MessageSensor(
                channel_id=1,
                sensor_type=SensorType.ENERGY_METER_VOLTAGE,
                value=Decimal(str(self.voltage_ch1)),
            ),
        )

        if self.voltage_ch2 and self.voltage_ch3:
            sensors.extend(
                [
                    IntegrationV0MessageSensor(
                        channel_id=1,
                        sensor_type=SensorType.ENERGY_METER_VOLTAGE,
                        value=Decimal(str(self.voltage_ch2)),
                    ),
                    IntegrationV0MessageSensor(
                        channel_id=1,
                        sensor_type=SensorType.ENERGY_METER_VOLTAGE,
                        value=Decimal(str(self.voltage_ch3)),
                    ),
                ],
            )

        sensors.append(
            IntegrationV0MessageSensor(
                channel_id=1,
                sensor_type=SensorType.ENERGY_METER_POWER_FACTOR,
                value=Decimal(str(self.k_ch1)),
            ),
        )

        if self.k_ch2 and self.k_ch3:
            sensors.extend(
                [
                    IntegrationV0MessageSensor(
                        channel_id=1,
                        sensor_type=SensorType.ENERGY_METER_POWER_FACTOR,
                        value=Decimal(str(self.k_ch2)),
                    ),
                    IntegrationV0MessageSensor(
                        channel_id=1,
                        sensor_type=SensorType.ENERGY_METER_POWER_FACTOR,
                        value=Decimal(str(self.k_ch3)),
                    ),
                ],
            )

        sensors.append(
            IntegrationV0MessageSensor(
                channel_id=1,
                sensor_type=SensorType.ENERGY_METER_FREQUENCY,
                value=Decimal(str(self.freq)),
            ),
        )

        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=sensors,
            ),
        ]



from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter import SmpmUlDeviceJupiter08BCounterData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_jupiter_08b_counter() -> None:
    case_serialized = bytes.fromhex("65fffffffbffffdf")
    assert SmpmUlDeviceJupiter08BCounterData(value_channel_1=67108863, value_channel_2=67108863, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterData.serialize(SmpmUlDeviceJupiter08BCounterData(value_channel_1=67108863, value_channel_2=67108863, event_reset=True, event_low_battery_level=True))  # noqa: E501
    case_serialized = bytes.fromhex("65000000000000c0")  # overflow
    assert SmpmUlDeviceJupiter08BCounterData(value_channel_1=0, value_channel_2=0, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterData.serialize(SmpmUlDeviceJupiter08BCounterData(event_low_battery_level=True, event_reset=True, value_channel_1=0, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("65000000000000c0")  # overflow
    assert SmpmUlDeviceJupiter08BCounterData(value_channel_1=0, value_channel_2=0, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterData.serialize(SmpmUlDeviceJupiter08BCounterData(event_low_battery_level=True, event_reset=True, value_channel_1=0, value_channel_2=268435456))  # noqa: E501
    case_serialized = bytes.fromhex("65000000000000c0")  # overflow
    assert SmpmUlDeviceJupiter08BCounterData(value_channel_1=0, value_channel_2=0, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterData.serialize(SmpmUlDeviceJupiter08BCounterData(event_low_battery_level=True, event_reset=True, value_channel_1=268435456, value_channel_2=0))  # noqa: E501
    case_serialized = bytes.fromhex("65000000000000c0")  # overflow
    assert SmpmUlDeviceJupiter08BCounterData(value_channel_1=0, value_channel_2=0, event_reset=True, event_low_battery_level=True) == SmpmUlDeviceJupiter08BCounterData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceJupiter08BCounterData.serialize(SmpmUlDeviceJupiter08BCounterData(event_low_battery_level=True, event_reset=True, value_channel_1=268435456, value_channel_2=268435456))  # noqa: E501



from data_gateway_sdk.device_data_encryption import EncryptionType


def test_decrypt_xtea_v_nero_v0() -> None:
    assert EncryptionType.XTEA_V_NERO_V0.decrypt(bytes.fromhex('8725024c66319b82'), bytes.fromhex('19208f5fd151c0b7e7c7890610f994b5c4fe697620b67e509b0ea2bd942b1fd6')) == bytes.fromhex('73086e0a000000a5')  # noqa: W391
    assert EncryptionType.XTEA_V_NERO_V0.decrypt(bytes.fromhex('461dee8050910076'), bytes.fromhex('6b4095ae9e35c7f4cbbd31de33a023e3f2d01d7c6041a885a13d202643118812')) == bytes.fromhex('7b3a00000000001f')  # noqa: W391
    assert EncryptionType.XTEA_V_NERO_V0.decrypt(bytes.fromhex('1ff19c361fd435cb'), bytes.fromhex('21dfdaae138b13c4c40fa2a560e51cbeca849f5b0d5b02232c27559125ddb124')) == bytes.fromhex('54c2000000000000')  # noqa: W391


def test_encrypt_xtea_v_nero_v0() -> None:
    assert EncryptionType.XTEA_V_NERO_V0.encrypt(bytes.fromhex('73086e0a000000a5'), bytes.fromhex('19208f5fd151c0b7e7c7890610f994b5c4fe697620b67e509b0ea2bd942b1fd6')) == bytes.fromhex('8725024c66319b82')  # noqa: W391
    assert EncryptionType.XTEA_V_NERO_V0.encrypt(bytes.fromhex('7b3a00000000001f'), bytes.fromhex('6b4095ae9e35c7f4cbbd31de33a023e3f2d01d7c6041a885a13d202643118812')) == bytes.fromhex('461dee8050910076')  # noqa: W391
    assert EncryptionType.XTEA_V_NERO_V0.encrypt(bytes.fromhex('54c2000000000000'), bytes.fromhex('21dfdaae138b13c4c40fa2a560e51cbeca849f5b0d5b02232c27559125ddb124')) == bytes.fromhex('1ff19c361fd435cb')  # noqa: W391



import json
import os.path
from datetime import datetime, timezone

from data_gateway_sdk.nero_bs_protocol import NeroBsProtocolType
from data_gateway_sdk.protocols.nero_bs_packet.http_nero_bs_packet import HttpV0NeroBsPacket, BaseStationInfo, BaseStationInfoGeo, BaseStationAdditionalInfo, DevicePackage, \
    DataNbfi, SdrInfo, DevicePackageProtocolTypeEnum, DevicePackageNetworkTypeEnum, SdrConfig, SdrConfigData, DModEnum, BaseStationLogs, BaseStationLogsTypeEnum

with open(os.path.join(os.path.dirname(__file__), 'bs_http_data.json')) as valid_json_f:
    VALID_BS_HTTP_DATA = json.load(valid_json_f)


def test_parsing_valid_bs_http() -> None:
    assert NeroBsProtocolType.NERO_BS_HTTP_V0.parse(VALID_BS_HTTP_DATA) == HttpV0NeroBsPacket(
        data=[
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 29, 1, tzinfo=timezone.utc),
                raw='000000008b6a3672fa335e6eac5ab668',
                data_nbfi=DataNbfi(mac=0, f_ask=False, iterator=0, multi=False, system=False, payload='8b6a3672fa335e6e'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868792300, freq_channel=89, sdr=0, baud_rate=50, rssi=-65, snr=67),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11986,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 29, 5, tzinfo=timezone.utc),
                raw='000000008b6a3672fa335e6eac5ab668',
                data_nbfi=DataNbfi(mac=0, f_ask=False, iterator=0, multi=False, system=False, payload='8b6a3672fa335e6e'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868819900, freq_channel=227, sdr=0, baud_rate=50, rssi=-68, snr=60),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11987,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 29, 16, tzinfo=timezone.utc),
                raw='8007a40dcf536171e112d18077342161',
                data_nbfi=DataNbfi(mac=8390564, f_ask=False, iterator=13, multi=False, system=False, payload='cf536171e112d180'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868796500, freq_channel=110, sdr=0, baud_rate=50, rssi=-91, snr=24),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11988,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 29, 23, tzinfo=timezone.utc),
                raw='8007a40dcf536171e112d18077342161',
                data_nbfi=DataNbfi(mac=8390564, f_ask=False, iterator=13, multi=False, system=False, payload='cf536171e112d180'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868823900, freq_channel=247, sdr=0, baud_rate=50, rssi=-93, snr=19),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11989,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 30, 59, tzinfo=timezone.utc),
                raw='802f0f00f54cdc01cdbe45a6a559ce7f',
                data_nbfi=DataNbfi(mac=8400655, f_ask=False, iterator=0, multi=False, system=False, payload='f54cdc01cdbe45a6'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868793300, freq_channel=94, sdr=0, baud_rate=50, rssi=-93, snr=39),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11990,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 31, 7, tzinfo=timezone.utc),
                raw='802f0f00f54cdc01cdbe45a6a559ce7f',
                data_nbfi=DataNbfi(mac=8400655, f_ask=False, iterator=0, multi=False, system=False, payload='f54cdc01cdbe45a6'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868820700, freq_channel=231, sdr=0, baud_rate=50, rssi=-99, snr=33),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11991,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 31, 20, tzinfo=timezone.utc),
                raw='8029740acf1586ec0de61a9c9d92105a',
                data_nbfi=DataNbfi(mac=8399220, f_ask=False, iterator=10, multi=False, system=False, payload='cf1586ec0de61a9c'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868784300, freq_channel=49, sdr=0, baud_rate=50, rssi=-66, snr=62),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11992,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 31, 28, tzinfo=timezone.utc), raw='8029740acf1586ec0de61a9c9d92105a',
                data_nbfi=DataNbfi(mac=8399220, f_ask=False, iterator=10, multi=False, system=False, payload='cf1586ec0de61a9c'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868812100, freq_channel=188, sdr=0, baud_rate=50, rssi=-68, snr=52),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11993,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 31, 57, tzinfo=timezone.utc), raw='80d19006f13ebae6a9c7b66f7d2a87f1',
                data_nbfi=DataNbfi(mac=8442256, f_ask=False, iterator=6, multi=False, system=False, payload='f13ebae6a9c7b66f'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868809100, freq_channel=173, sdr=0, baud_rate=50, rssi=-97, snr=15),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11994,  # type: ignore
            ),
            DevicePackage(
                dt=datetime(2023, 2, 1, 13, 31, 59, tzinfo=timezone.utc),
                raw='802f0f009042a07fc08a377e1dba7f62',
                data_nbfi=DataNbfi(mac=8400655, f_ask=False, iterator=0, multi=False, system=False, payload='9042a07fc08a377e'),
                data_unbp=None,
                data_unbp2=None,
                sdr_info=SdrInfo(freq=868779500, freq_channel=25, sdr=0, baud_rate=50, rssi=-98, snr=16),
                type=DevicePackageNetworkTypeEnum.uplink,
                protocol=DevicePackageProtocolTypeEnum.nbfi,
                message_id=None,
                id=11995,  # type: ignore
            ),
        ],
        base_station_environment_info=[],
        base_station_info=BaseStationInfo(
            geo=BaseStationInfoGeo(dt=datetime(2023, 2, 2, 15, 23, 15, tzinfo=timezone.utc), longitude=0.0, latitude=0.0, actual=False),
            uptime_s=247767,
            sdr_config=SdrConfig(
                freq=868800000,
                sdr=[
                    SdrConfigData(enable=True, dds_freq=0, dmod=DModEnum.DBPSK, baud_rate=50, unions=4),
                    SdrConfigData(enable=True, dds_freq=0, dmod=DModEnum.DBPSK, baud_rate=100, unions=4),
                ],
            ),
        ),
        base_station_additional_info=BaseStationAdditionalInfo(id=9017, version='1.6.0-3-17.12.2022'),
        base_station_logs=[
            BaseStationLogs(type=BaseStationLogsTypeEnum.warning, mdl='lte', dt=datetime(2023, 2, 1, 13, 28, 47, tzinfo=timezone.utc), text='GNSS Info NOT received'),  # noqa: E501
            BaseStationLogs(type=BaseStationLogsTypeEnum.warning, mdl='lte', dt=datetime(2023, 2, 1, 13, 28, 58, tzinfo=timezone.utc), text='GNSS Info NOT received'),  # noqa: E501
            BaseStationLogs(type=BaseStationLogsTypeEnum.warning, mdl='lte', dt=datetime(2023, 2, 1, 13, 29, 9, tzinfo=timezone.utc), text='GNSS Info NOT received'),
            BaseStationLogs(type=BaseStationLogsTypeEnum.warning, mdl='lte', dt=datetime(2023, 2, 1, 13, 29, 20, tzinfo=timezone.utc), text='GNSS Info NOT received'),  # noqa: E501
            BaseStationLogs(type=BaseStationLogsTypeEnum.warning, mdl='lte', dt=datetime(2023, 2, 1, 13, 29, 31, tzinfo=timezone.utc), text='GNSS Info NOT received'),  # noqa: E501
            BaseStationLogs(type=BaseStationLogsTypeEnum.warning, mdl='lte', dt=datetime(2023, 2, 1, 13, 29, 42, tzinfo=timezone.utc), text='GNSS Info NOT received'),  # noqa: E501
            BaseStationLogs(type=BaseStationLogsTypeEnum.warning, mdl='lte', dt=datetime(2023, 2, 1, 13, 29, 53, tzinfo=timezone.utc), text='GNSS Info NOT received'),  # noqa: E501
            BaseStationLogs(type=BaseStationLogsTypeEnum.warning, mdl='lte', dt=datetime(2023, 2, 1, 13, 30, 4, tzinfo=timezone.utc), text='GNSS Info NOT received'),
        ],
    )





from typing import List, Tuple

from data_gateway_sdk.crypto_algorithms.extensions.errors import XteaError

delta: int = 2654435769
num_iterations = 64
key_length_bytes = 32
block_length_bytes = 8
mask = 0xffffffff


def decrypt(src: str, key_bytes: bytes) -> bytes:
    assert isinstance(src, str), f'src decrypt must be type of {str}, got type {type(src)} '
    assert isinstance(key_bytes, bytes), f'key_bytes decrypt must be type of {bytes}, got type {type(key_bytes)} '
    data = bytearray.fromhex(src)

    valid_key(key_bytes)
    tmp_decrypts = _decrypt_xtea(data, key_bytes[16:32])
    return _decrypt_xtea(tmp_decrypts, key_bytes[0:16])


def xtea_decrypt(src: bytes, key_bytes: bytes) -> bytes:
    assert isinstance(src, bytes), f'src decrypt must be type of {bytes}, got type {type(src)} '
    assert isinstance(key_bytes, bytes), f'key_bytes decrypt must be type of {bytes}, got type {type(key_bytes)} '
    valid_key(key_bytes)
    tmp_decrypts = _decrypt_xtea(src, key_bytes[16:32])

    return _decrypt_xtea(tmp_decrypts, key_bytes[0:16])


def encrypt(src: str, key_bytes: bytes) -> bytes:
    assert isinstance(src, str), f'src encrypt must be type of {str}, got type {type(src)} '
    assert isinstance(key_bytes, bytes), f'key_bytes encrypt must be type of {bytes}, got type {type(key_bytes)} '
    data = bytearray.fromhex(src)

    valid_key(key_bytes)
    tmp_decrypts = _encrypt_xtea(data, key_bytes[0:16])
    return _encrypt_xtea(tmp_decrypts, key_bytes[16:32])


def xtea_encrypt(src: bytes, key_bytes: bytes) -> bytes:
    assert isinstance(src, bytes), f'src decrypt must be type of {bytes}, got type {type(src)} '
    assert isinstance(key_bytes, bytes), f'key_bytes decrypt must be type of {bytes}, got type {type(key_bytes)} '

    valid_key(key_bytes)
    tmp_decrypts = _encrypt_xtea(src, key_bytes[0:16])
    return _encrypt_xtea(tmp_decrypts, key_bytes[16:32])


def _decrypt_xtea(src: bytes, key_bytes: bytes) -> bytes:
    dst: bytes = b""
    blocks_num, key = _preparation_block(src, key_bytes)
    for i in range(blocks_num):
        block_start = i * block_length_bytes
        dst += _decrypt_block(src[block_start:block_start + block_length_bytes], key)
    return dst


def _encrypt_xtea(src: bytes, key_bytes: bytes) -> bytes:
    dst: bytes = b""
    blocks_num, key = _preparation_block(src, key_bytes)
    for i in range(blocks_num):
        block_start = i * block_length_bytes
        dst += _encrypt_block(src[block_start:block_start + block_length_bytes], key)
    return dst


def _preparation_block(src: bytes, key_bytes: bytes) -> Tuple[int, List[int]]:
    key: List[int] = []
    key.insert(0, (int.from_bytes(bytes=key_bytes[0:4], byteorder="little")))
    key.insert(1, (int.from_bytes(bytes=key_bytes[4:8], byteorder="little")))
    key.insert(2, (int.from_bytes(bytes=key_bytes[8:12], byteorder="little")))
    key.insert(3, (int.from_bytes(bytes=key_bytes[12:16], byteorder="little")))

    blocks_num: int = int(len(src) / block_length_bytes)
    return blocks_num, key


def _decrypt_block(src: bytes, key: List[int]) -> bytes:
    amount = (delta * num_iterations) & mask
    x1, x2 = _block_to_int32(src)
    for _ in range(num_iterations):
        x2 = (x2 - (((x1 << 4 ^ x1 >> 5) + x1) ^ (amount + key[(amount >> 11) & 3])) & mask)
        amount = (amount - delta) & mask
        x1 = (x1 - (((x2 << 4 ^ x2 >> 5) + x2) ^ (amount + key[amount & 3])) & mask)
    result = _uint32_to_block(x1, x2)
    return result


def _encrypt_block(src: bytes, key: List[int]) -> bytes:
    amount = 0
    x1, x2 = _block_to_int32(src)
    for _ in range(num_iterations):
        x1 = (x1 + (((x2 << 4 ^ x2 >> 5) + x2) ^ (amount + key[amount & 3]))) & mask
        amount = (amount + delta) & mask
        x2 = (x2 + (((x1 << 4 ^ x1 >> 5) + x1) ^ (amount + key[amount >> 11 & 3]))) & mask
    result = _uint32_to_block(x1, x2)
    return result


def _block_to_int32(block: bytes) -> Tuple[int, int]:
    x1 = int.from_bytes(block[0:4], "little")
    x2 = int.from_bytes(block[4:8], "little")
    return x1, x2


def _uint32_to_block(x1: int, x2: int) -> bytes:
    return x1.to_bytes(length=4, byteorder="little") + x2.to_bytes(length=4, byteorder="little")


def valid_key(key_bytes: bytes) -> None:
    if len(key_bytes) != key_length_bytes:
        raise XteaError(f"invalid length KEY '{len(key_bytes)}' was given, must '{key_length_bytes}'")



from datetime import timedelta, datetime, time, tzinfo
from decimal import Decimal
from typing import List, Any, Dict

from data_aggregator_sdk.constants.enums import JournalDataType, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageEvent, IntegrationV0MessageData, \
    IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageClock, IntegrationV0MessageSensor, \
    IntegrationV0MessageConnectionQuality

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet
from data_gateway_sdk.utils.timestamp_calculation import timestamp_calculation


# PACKET (128 bits)   smpm_ul_device_gaz_flow_16b_daily
#
# RESULT int:        10692673142131833941215815354860702850
# RESULT bin:  MSB   00001000 00001011 01010101 01111101 00000110 10110001 11101010 11100100 10010000 10111010 01111111 11111111 11111111 11000000 00001100 10000010   LSB
# RESULT hex:  LE    82 0c c0 ff ff 7f ba 90 e4 ea b1 06 7d 55 0b 08
#
# name                            type       size  value(int)                                                                                                                        data(bits)
# ------------------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL            u7            7           2                                                                                                                           0000010
# packet_type_id.0.DFF            bool          1           1                                                                                                                          1
# packet_type_id.1.VAL            u2            2           0                                                                                                                        00
# packet_type_id.1.DFF            bool          1           1                                                                                                                       1
# packet_type_id.2.VAL            u2            2           1                                                                                                                     01
# packet_type_id.2.DFF            bool          1           0                                                                                                                    0
# days_ago                        timedelta     5           0                                                                                                               00000
# sync_time_days_ago              timedelta     3           0                                                                                                            000
# timestamp_s                     timedelta    26    33554431                                                                                  01111111111111111111111111
# temperature                     u7            7          58                                                                           0111010
# battery_volts                   uf6p1         6          33                                                                     100001
# event_reset                     bool          1           0                                                                    0
# event_low_battery_level         bool          1           0                                                                   0
# event_temperature_limits        bool          1           1                                                                  1
# direct_flow_volume              uf32p3       32   112323300                                  00000110101100011110101011100100
# rsrp                            u7            7         125                           1111101
# rsrq                            u4            4          10                       1010
# snr                             u6            6          42                 101010
# ecl                             u2            2           1               01
# event_battery_warn              bool          1           1              1
# event_system_error              bool          1           0             0
# event_flow_reverse              bool          1           0            0
# event_flow_speed_is_over_limit  bool          1           0           0
# event_sensor_error              bool          1           0          0
# event_sensor_error_temperature  bool          1           0         0
# event_case_was_opened           bool          1           0        0
# event_continuous_consumption    bool          1           0       0
# event_no_resource               bool          1           1      1
# RESERVED                        u4            4           0  0000


class SmpmUlDeviceGazFlow16BDailyData(Packet):
    days_ago: timedelta
    sync_time_days_ago: timedelta
    timestamp_s: timedelta
    temperature: int
    battery_volts: float
    event_reset: bool
    event_low_battery_level: bool
    event_temperature_limits: bool
    direct_flow_volume: float
    rsrp: int
    rsrq: int
    snr: int
    ecl: int
    event_battery_warn: bool
    event_system_error: bool
    event_flow_reverse: bool
    event_flow_speed_is_over_limit: bool
    event_sensor_error: bool
    event_sensor_error_temperature: bool
    event_case_was_opened: bool
    event_continuous_consumption: bool
    event_no_resource: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((2) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 31
        result |= ((days_ago_tmp1) & (2 ** (5) - 1)) << size
        size += 5
        isinstance(data.sync_time_days_ago, (int, timedelta))
        result |= ((max(min(7, int(data.sync_time_days_ago.total_seconds() // 86400 if isinstance(data.sync_time_days_ago, timedelta) else data.sync_time_days_ago // 86400)), 0)) & (2 ** (3) - 1)) << size
        size += 3
        isinstance(data.timestamp_s, (int, timedelta))
        value_int_tmp2 = int(data.timestamp_s.total_seconds() // 1 if isinstance(data.timestamp_s, timedelta) else data.timestamp_s // 1) & 67108863
        result |= ((value_int_tmp2) & (2 ** (26) - 1)) << size
        size += 26
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 6.3
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_temperature_limits, bool)
        result |= ((int(data.event_temperature_limits)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.direct_flow_volume, (int, float))
        result |= ((int(round(float(data.direct_flow_volume) * 1000.0, 0)) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.rsrp, int)
        assert -140 <= data.rsrp <= -13
        result |= (((data.rsrp + 140)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.rsrq, int)
        assert -19 <= data.rsrq <= -4
        result |= (((data.rsrq + 19)) & (2 ** (4) - 1)) << size
        size += 4
        assert isinstance(data.snr, int)
        assert -20 <= data.snr <= 43
        result |= (((data.snr + 20)) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.ecl, int)
        assert 0 <= data.ecl <= 3
        result |= ((data.ecl) & (2 ** (2) - 1)) << size
        size += 2
        assert isinstance(data.event_battery_warn, bool)
        result |= ((int(data.event_battery_warn)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_system_error, bool)
        result |= ((int(data.event_system_error)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_flow_reverse, bool)
        result |= ((int(data.event_flow_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_flow_speed_is_over_limit, bool)
        result |= ((int(data.event_flow_speed_is_over_limit)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_sensor_error, bool)
        result |= ((int(data.event_sensor_error)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_sensor_error_temperature, bool)
        result |= ((int(data.event_sensor_error_temperature)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_case_was_opened, bool)
        result |= ((int(data.event_case_was_opened)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_continuous_consumption, bool)
        result |= ((int(data.event_continuous_consumption)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_no_resource, bool)
        result |= ((int(data.event_no_resource)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceGazFlow16BDailyData':
        result__el_tmp3: Dict[str, Any] = dict()
        if 2 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp3["days_ago"] = timedelta(seconds=buf.shift(5) * 86400)
        result__el_tmp3["sync_time_days_ago"] = timedelta(seconds=buf.shift(3) * 86400)
        result__el_tmp3["timestamp_s"] = timedelta(seconds=buf.shift(26) * 1)
        result__el_tmp3["temperature"] = buf.shift(7) + -35
        result__el_tmp3["battery_volts"] = round(buf.shift(6) / 10.0, 1)
        result__el_tmp3["event_reset"] = bool(buf.shift(1))
        result__el_tmp3["event_low_battery_level"] = bool(buf.shift(1))
        result__el_tmp3["event_temperature_limits"] = bool(buf.shift(1))
        result__el_tmp3["direct_flow_volume"] = round(buf.shift(32) / 1000.0, 3)
        result__el_tmp3["rsrp"] = buf.shift(7) + -140
        result__el_tmp3["rsrq"] = buf.shift(4) + -19
        result__el_tmp3["snr"] = buf.shift(6) + -20
        result__el_tmp3["ecl"] = buf.shift(2) + 0
        result__el_tmp3["event_battery_warn"] = bool(buf.shift(1))
        result__el_tmp3["event_system_error"] = bool(buf.shift(1))
        result__el_tmp3["event_flow_reverse"] = bool(buf.shift(1))
        result__el_tmp3["event_flow_speed_is_over_limit"] = bool(buf.shift(1))
        result__el_tmp3["event_sensor_error"] = bool(buf.shift(1))
        result__el_tmp3["event_sensor_error_temperature"] = bool(buf.shift(1))
        result__el_tmp3["event_case_was_opened"] = bool(buf.shift(1))
        result__el_tmp3["event_continuous_consumption"] = bool(buf.shift(1))
        result__el_tmp3["event_no_resource"] = bool(buf.shift(1))
        result = SmpmUlDeviceGazFlow16BDailyData(**result__el_tmp3)
        buf.shift(4)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=days_ago_calculation(received_at, device_tz, time(0), self.days_ago),
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.direct_flow_volume)),
                        overloading_value=Decimal(str(4294967.295)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.TEMPERATURE_LIMIT] if self.event_temperature_limits else []),
                    *([IntegrationV0MessageEvent.BATTERY_WARNING] if self.event_battery_warn else []),
                    *([IntegrationV0MessageEvent.ERROR_SYSTEM] if self.event_system_error else []),
                    *([IntegrationV0MessageEvent.FLOW_REVERSE] if self.event_flow_reverse else []),
                    *([IntegrationV0MessageEvent.FLOW_SPEED_OVER_LIMIT] if self.event_flow_speed_is_over_limit else []),
                    *([IntegrationV0MessageEvent.ERROR_SENSOR] if self.event_sensor_error else []),
                    *([IntegrationV0MessageEvent.ERROR_SENSOR_TEMPERATURE] if self.event_sensor_error_temperature else []),
                    *([IntegrationV0MessageEvent.CASE_WAS_OPENED] if self.event_case_was_opened else []),
                    *([IntegrationV0MessageEvent.CONTINUES_CONSUMPTION] if self.event_continuous_consumption else []),
                    *([IntegrationV0MessageEvent.NO_RESOURCE] if self.event_no_resource else []),
                ],
                clock=[IntegrationV0MessageClock(value=timestamp_calculation(received_at, self.timestamp_s, timedelta(seconds=67108863)))] if self.timestamp_s else [],
                connection_quality=IntegrationV0MessageConnectionQuality(
                    rsrp=self.rsrp,
                    rsrq=self.rsrq,
                    snr=self.snr,
                    ecl=self.ecl,
                )
            ),
        ]



from datetime import timedelta, datetime, time, tzinfo
from enum import IntEnum, unique
from typing import List, Any, Dict, Tuple, Optional

from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageProfile, ProfileGranulation, ProfileKind, GRANULATION_TO_END_OF_DATETIME_MAP
from data_aggregator_sdk.utils.round_dt import round_dt

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet
from data_gateway_sdk.utils.true_round import true_round


# PACKET (128 bits)   smpm_ul_device_energy_16b_profile_8h_f
#
# RESULT int:        237762635927079402885851130105524
# RESULT bin:  MSB   00000000 00000000 00001011 10111000 11111100 10000010 11010100 00010000 11111010 00010000 01000000 00010100 00000001 11100000 00001110 10110100   LSB
# RESULT hex:  LE    b4 0e e0 01 14 40 10 fa 10 d4 82 fc b8 0b 00 00
#
# name                  type       size  value(int)                                                                                                                        data(bits)   # noqa: E501
# --------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL  u7            7          52                                                                                                                           0110100   # noqa: E501
# packet_type_id.0.DFF  bool          1           1                                                                                                                          1
# packet_type_id.1.VAL  u2            2           2                                                                                                                        10
# packet_type_id.1.DFF  bool          1           1                                                                                                                       1
# packet_type_id.2.VAL  u2            2           1                                                                                                                     01
# packet_type_id.2.DFF  bool          1           0                                                                                                                    0
# days_ago              timedelta     6           0                                                                                                              000000
# profile.0.point       uf11p2       11          30                                                                                                   00000011110
# profile.1.point       uf11p2       11          40                                                                                        00000101000
# profile.2.point       uf11p2       11        1040                                                                             10000010000
# profile.3.point       uf11p2       11        2000                                                                  11111010000
# profile.4.point       uf11p2       11        1040                                                       10000010000
# profile.5.point       uf11p2       11          90                                            00001011010
# profile.6.point       uf11p2       11        1010                                 01111110010
# profile.7.point       uf11p2       11        1500                      10111011100
# RESERVED              u20          20           0  00000000000000000000
@unique
class SmpmUlDeviceEnergy16bProfile8hFIds(IntEnum):
    UL_DATA_16B__PROFILE_H8_1_F_AVG = 820
    UL_DATA_16B__PROFILE_H8_1_F_MIN = 821
    UL_DATA_16B__PROFILE_H8_1_F_MAX = 822
    UL_DATA_16B__PROFILE_H8_2_F_AVG = 823
    UL_DATA_16B__PROFILE_H8_2_F_MIN = 824
    UL_DATA_16B__PROFILE_H8_2_F_MAX = 825
    UL_DATA_16B__PROFILE_H8_3_F_AVG = 826
    UL_DATA_16B__PROFILE_H8_3_F_MIN = 827
    UL_DATA_16B__PROFILE_H8_3_F_MAX = 828
    UL_DATA_16B__PROFILE_H8_1_F_AVG__PHASE_A = 829
    UL_DATA_16B__PROFILE_H8_1_F_MIN__PHASE_A = 830
    UL_DATA_16B__PROFILE_H8_1_F_MAX__PHASE_A = 831
    UL_DATA_16B__PROFILE_H8_2_F_AVG__PHASE_A = 832
    UL_DATA_16B__PROFILE_H8_2_F_MIN__PHASE_A = 833
    UL_DATA_16B__PROFILE_H8_2_F_MAX__PHASE_A = 834
    UL_DATA_16B__PROFILE_H8_3_F_AVG__PHASE_A = 835
    UL_DATA_16B__PROFILE_H8_3_F_MIN__PHASE_A = 836
    UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_A = 837
    UL_DATA_16B__PROFILE_H8_1_F_AVG__PHASE_B = 838
    UL_DATA_16B__PROFILE_H8_1_F_MIN__PHASE_B = 839
    UL_DATA_16B__PROFILE_H8_1_F_MAX__PHASE_B = 840
    UL_DATA_16B__PROFILE_H8_2_F_AVG__PHASE_B = 841
    UL_DATA_16B__PROFILE_H8_2_F_MIN__PHASE_B = 842
    UL_DATA_16B__PROFILE_H8_2_F_MAX__PHASE_B = 843
    UL_DATA_16B__PROFILE_H8_3_F_AVG__PHASE_B = 844
    UL_DATA_16B__PROFILE_H8_3_F_MIN__PHASE_B = 845
    UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_B = 846
    UL_DATA_16B__PROFILE_H8_1_F_AVG__PHASE_C = 847
    UL_DATA_16B__PROFILE_H8_1_F_MIN__PHASE_C = 848
    UL_DATA_16B__PROFILE_H8_1_F_MAX__PHASE_C = 849
    UL_DATA_16B__PROFILE_H8_2_F_AVG__PHASE_C = 850
    UL_DATA_16B__PROFILE_H8_2_F_MIN__PHASE_C = 851
    UL_DATA_16B__PROFILE_H8_2_F_MAX__PHASE_C = 852
    UL_DATA_16B__PROFILE_H8_3_F_AVG__PHASE_C = 853
    UL_DATA_16B__PROFILE_H8_3_F_MIN__PHASE_C = 854
    UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_C = 855


PROFILE_H8_F_TYPE_MAP = {
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG: ProfileKind.FREQUENCY_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_MIN: ProfileKind.FREQUENCY_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_MAX: ProfileKind.FREQUENCY_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_AVG: ProfileKind.FREQUENCY_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_MIN: ProfileKind.FREQUENCY_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_MAX: ProfileKind.FREQUENCY_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_AVG: ProfileKind.FREQUENCY_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MIN: ProfileKind.FREQUENCY_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MAX: ProfileKind.FREQUENCY_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG__PHASE_A: ProfileKind.FREQUENCY_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_MIN__PHASE_A: ProfileKind.FREQUENCY_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_MAX__PHASE_A: ProfileKind.FREQUENCY_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_AVG__PHASE_A: ProfileKind.FREQUENCY_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_MIN__PHASE_A: ProfileKind.FREQUENCY_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_MAX__PHASE_A: ProfileKind.FREQUENCY_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_AVG__PHASE_A: ProfileKind.FREQUENCY_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MIN__PHASE_A: ProfileKind.FREQUENCY_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_A: ProfileKind.FREQUENCY_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG__PHASE_B: ProfileKind.FREQUENCY_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_MIN__PHASE_B: ProfileKind.FREQUENCY_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_MAX__PHASE_B: ProfileKind.FREQUENCY_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_AVG__PHASE_B: ProfileKind.FREQUENCY_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_MIN__PHASE_B: ProfileKind.FREQUENCY_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_MAX__PHASE_B: ProfileKind.FREQUENCY_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_AVG__PHASE_B: ProfileKind.FREQUENCY_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MIN__PHASE_B: ProfileKind.FREQUENCY_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_B: ProfileKind.FREQUENCY_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG__PHASE_C: ProfileKind.FREQUENCY_C,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_MIN__PHASE_C: ProfileKind.FREQUENCY_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_MAX__PHASE_C: ProfileKind.FREQUENCY_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_AVG__PHASE_C: ProfileKind.FREQUENCY_C,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_MIN__PHASE_C: ProfileKind.FREQUENCY_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_2_F_MAX__PHASE_C: ProfileKind.FREQUENCY_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_AVG__PHASE_C: ProfileKind.FREQUENCY_C,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MIN__PHASE_C: ProfileKind.FREQUENCY_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_C: ProfileKind.FREQUENCY_MAX_C,
}

assert set(PROFILE_H8_F_TYPE_MAP.keys()) == set(v for v in SmpmUlDeviceEnergy16bProfile8hFIds)


class SmpmUlDeviceEnergy16BProfile8HFData(Packet):
    packet_type_id: SmpmUlDeviceEnergy16bProfile8hFIds
    days_ago: timedelta
    profile: Tuple[float, float, float, float, float, float, float, float]

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        packet_type_id__value_int_tmp1 = 0
        packet_type_id__value_size_tmp2 = 0
        assert isinstance(data.packet_type_id, SmpmUlDeviceEnergy16bProfile8hFIds)
        packet_type_id__value_int_tmp1 |= ((data.packet_type_id.value) & (2 ** (10) - 1)) << packet_type_id__value_size_tmp2
        packet_type_id__value_size_tmp2 += 10
        packet_type_id__max_size_tmp3 = 0
        packet_type_id__steps_tmp4 = [8, 3, 3, 3]
        for packet_type_id__j_tmp5 in range(32):
            packet_type_id__step_tmp6 = packet_type_id__steps_tmp4[packet_type_id__j_tmp5] if packet_type_id__j_tmp5 < len(packet_type_id__steps_tmp4) else packet_type_id__steps_tmp4[-1]  # noqa: E501
            packet_type_id__max_size_tmp3 += packet_type_id__step_tmp6
            packet_type_id__current_part_value_tmp7 = packet_type_id__value_int_tmp1 & (2 ** packet_type_id__step_tmp6 - 1)
            packet_type_id__value_int_tmp1 = packet_type_id__value_int_tmp1 >> (packet_type_id__step_tmp6 - 1)
            result |= ((packet_type_id__current_part_value_tmp7) & (2 ** ((packet_type_id__step_tmp6 - 1)) - 1)) << size
            size += (packet_type_id__step_tmp6 - 1)
            assert isinstance((packet_type_id__value_int_tmp1 != 0), bool)
            result |= ((int((packet_type_id__value_int_tmp1 != 0))) & (2 ** (1) - 1)) << size
            size += 1
            if packet_type_id__value_int_tmp1 == 0:
                break
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp8 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp8 <= 63
        result |= ((days_ago_tmp8) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.profile, tuple) and len(data.profile) == 8
        assert isinstance(data.profile[0], (int, float))
        assert 45.0 <= data.profile[0] <= 65.47
        result |= ((int(round(float(data.profile[0] - 45.0) * 100.0, 0))) & (2 ** (11) - 1)) << size
        size += 11
        assert isinstance(data.profile[1], (int, float))
        assert 45.0 <= data.profile[1] <= 65.47
        result |= ((int(round(float(data.profile[1] - 45.0) * 100.0, 0))) & (2 ** (11) - 1)) << size
        size += 11
        assert isinstance(data.profile[2], (int, float))
        assert 45.0 <= data.profile[2] <= 65.47
        result |= ((int(round(float(data.profile[2] - 45.0) * 100.0, 0))) & (2 ** (11) - 1)) << size
        size += 11
        assert isinstance(data.profile[3], (int, float))
        assert 45.0 <= data.profile[3] <= 65.47
        result |= ((int(round(float(data.profile[3] - 45.0) * 100.0, 0))) & (2 ** (11) - 1)) << size
        size += 11
        assert isinstance(data.profile[4], (int, float))
        assert 45.0 <= data.profile[4] <= 65.47
        result |= ((int(round(float(data.profile[4] - 45.0) * 100.0, 0))) & (2 ** (11) - 1)) << size
        size += 11
        assert isinstance(data.profile[5], (int, float))
        assert 45.0 <= data.profile[5] <= 65.47
        result |= ((int(round(float(data.profile[5] - 45.0) * 100.0, 0))) & (2 ** (11) - 1)) << size
        size += 11
        assert isinstance(data.profile[6], (int, float))
        assert 45.0 <= data.profile[6] <= 65.47
        result |= ((int(round(float(data.profile[6] - 45.0) * 100.0, 0))) & (2 ** (11) - 1)) << size
        size += 11
        assert isinstance(data.profile[7], (int, float))
        assert 45.0 <= data.profile[7] <= 65.47
        result |= ((int(round(float(data.profile[7] - 45.0) * 100.0, 0))) & (2 ** (11) - 1)) << size
        size += 11
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BProfile8HFData':
        result__el_tmp9: Dict[str, Any] = dict()
        packet_type_id__res_tmp10 = 0
        packet_type_id__steps_tmp11 = (8, 3, 3, 3)
        packet_type_id__res_size_tmp12 = 0
        packet_type_id__step_tmp13 = 0
        for packet_type_id__i_tmp15 in range(32):
            packet_type_id__step_tmp13 = (packet_type_id__steps_tmp11[packet_type_id__i_tmp15] if packet_type_id__i_tmp15 < len(packet_type_id__steps_tmp11) else packet_type_id__steps_tmp11[-1]) - 1  # noqa: E501
            packet_type_id__res_tmp10 |= buf.shift(packet_type_id__step_tmp13) << packet_type_id__res_size_tmp12
            packet_type_id__res_size_tmp12 += packet_type_id__step_tmp13
            packet_type_id__dff_tmp14 = bool(buf.shift(1))
            if not packet_type_id__dff_tmp14:
                break
        packet_type_id__buf_tmp16 = buf
        buf = BufRef(packet_type_id__res_tmp10, stop_on_buffer_end=True)
        result__el_tmp9["packet_type_id"] = SmpmUlDeviceEnergy16bProfile8hFIds(buf.shift(10))
        buf = packet_type_id__buf_tmp16
        result__el_tmp9["days_ago"] = timedelta(seconds=buf.shift(6) * 86400)
        profile_tmp17: List[float] = []
        profile__item_tmp18 = round(buf.shift(11) / 100.0 + 45.0, 2)
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = round(buf.shift(11) / 100.0 + 45.0, 2)
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = round(buf.shift(11) / 100.0 + 45.0, 2)
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = round(buf.shift(11) / 100.0 + 45.0, 2)
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = round(buf.shift(11) / 100.0 + 45.0, 2)
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = round(buf.shift(11) / 100.0 + 45.0, 2)
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = round(buf.shift(11) / 100.0 + 45.0, 2)
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = round(buf.shift(11) / 100.0 + 45.0, 2)
        profile_tmp17.append(profile__item_tmp18)
        result__el_tmp9["profile"] = tuple(profile_tmp17)
        result = SmpmUlDeviceEnergy16BProfile8HFData(**result__el_tmp9)
        buf.shift(20)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        values: Tuple[Optional[float], ...]
        precision = 2
        profiles = (
            true_round(self.profile[0], precision),
            true_round(self.profile[1], precision),
            true_round(self.profile[2], precision),
            true_round(self.profile[3], precision),
            true_round(self.profile[4], precision),
            true_round(self.profile[5], precision),
            true_round(self.profile[6], precision),
            true_round(self.profile[7], precision),
        )
        if self.packet_type_id in {820, 821, 822, 829, 830, 831, 838, 839, 840, 847, 848, 849}:
            values = (
                *profiles,
                None, None, None, None, None, None, None, None,
                None, None, None, None, None, None, None, None,
            )
        elif self.packet_type_id in {823, 824, 825, 832, 833, 834, 841, 842, 843, 850, 851, 852}:
            values = (
                None, None, None, None, None, None, None, None,
                *profiles,
                None, None, None, None, None, None, None, None,
            )
        else:
            values = (
                None, None, None, None, None, None, None, None,
                None, None, None, None, None, None, None, None,
                *profiles,
            )
        return [
            IntegrationV0MessageData(
                dt=round_dt(
                    days_ago_calculation(
                        received_at,
                        device_tz,
                        time(0),
                        self.days_ago,
                    ),
                    GRANULATION_TO_END_OF_DATETIME_MAP[ProfileGranulation.MINUTE_60],
                ),
                profiles=[
                    IntegrationV0MessageProfile(
                        type=PROFILE_H8_F_TYPE_MAP[self.packet_type_id],
                        granulation=ProfileGranulation.MINUTE_60,
                        values=values,
                    ),
                ],
            ),
        ]



from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_info import SmpmUlDeviceWaterMeter08BInfoData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_water_meter_08b_info() -> None:
    case_serialized = bytes.fromhex("69e14a1400000000")  # overflow
    assert SmpmUlDeviceWaterMeter08BInfoData(battery_volts=3.3, event_battery_low=True, event_battery_warn=True, reverse_flow_volume=10.98, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BInfoData.serialize(SmpmUlDeviceWaterMeter08BInfoData(battery_volts=3.3, event_battery_low=True, event_battery_warn=True, reverse_flow_volume=112323.3, event_flow_reverse=True))  # noqa: E501
    case_serialized = bytes.fromhex("69c0001000000000")  # overflow
    assert SmpmUlDeviceWaterMeter08BInfoData(battery_volts=0.0, event_battery_low=True, event_battery_warn=True, reverse_flow_volume=0.0, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BInfoData.serialize(SmpmUlDeviceWaterMeter08BInfoData(battery_volts=0.0, event_battery_low=True, event_battery_warn=True, event_flow_reverse=True, reverse_flow_volume=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("69c0001000000000")  # overflow
    assert SmpmUlDeviceWaterMeter08BInfoData(battery_volts=0.0, event_battery_low=True, event_battery_warn=True, reverse_flow_volume=0.0, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BInfoData.serialize(SmpmUlDeviceWaterMeter08BInfoData(battery_volts=0.0, event_battery_low=True, event_battery_warn=True, event_flow_reverse=True, reverse_flow_volume=81.92))  # noqa: E501
    case_serialized = bytes.fromhex("69ff001000000000")  # overflow
    assert SmpmUlDeviceWaterMeter08BInfoData(battery_volts=6.3, event_battery_low=True, event_battery_warn=True, reverse_flow_volume=0.0, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BInfoData.serialize(SmpmUlDeviceWaterMeter08BInfoData(battery_volts=6.3, event_battery_low=True, event_battery_warn=True, event_flow_reverse=True, reverse_flow_volume=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("69ff001000000000")  # overflow
    assert SmpmUlDeviceWaterMeter08BInfoData(battery_volts=6.3, event_battery_low=True, event_battery_warn=True, reverse_flow_volume=0.0, event_flow_reverse=True) == SmpmUlDeviceWaterMeter08BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceWaterMeter08BInfoData.serialize(SmpmUlDeviceWaterMeter08BInfoData(battery_volts=6.3, event_battery_low=True, event_battery_warn=True, event_flow_reverse=True, reverse_flow_volume=81.92))  # noqa: E501



from typing import Tuple, Type

from data_gateway_sdk.protocols.device_packet.device_packet import DevicePacketTyped
from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_16b_get_data import SmpmDlDeviceEnergy16BGetDataData
from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_16b_set_regular_data_sending import \
    SmpmDlDeviceEnergy16BSetRegularDataSendingData
from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_8b_get_data import SmpmDlDeviceEnergy8BGetDataData
from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_8b_set_clock import SmpmDlDeviceEnergy8BSetClockData
from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_8b_set_relay import SmpmDlDeviceEnergy8BSetRelayData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_dl_answer import SmpmUlDeviceDlAnswerData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_3phase_consumed import SmpmUlDeviceEnergy16B3PhaseConsumedData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_3phase_generated import \
    SmpmUlDeviceEnergy16B3PhaseGeneratedData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_daily import SmpmUlDeviceEnergy16BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_device_info import SmpmUlDeviceEnergy16BDeviceInfoData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_electricity_params import SmpmUlDeviceEnergy16BElectricityParamsData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_info import SmpmUlDeviceEnergy16BInfoData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_journal import SmpmUlDeviceEnergy16BJournalData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h1_energy import \
    SmpmUlDeviceEnergy16BProfile8H1EnergyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h2_energy import \
    SmpmUlDeviceEnergy16BProfile8H2EnergyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h3_energy import \
    SmpmUlDeviceEnergy16BProfile8H3EnergyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h_ar import SmpmUlDeviceEnergy16BProfile8HArData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h_energy import SmpmUlDeviceEnergy16BProfile8HEnergyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h_f import SmpmUlDeviceEnergy16BProfile8HFData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h_pqs import SmpmUlDeviceEnergy16BProfile8HPqsData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_retrospective_energy import SmpmUlDeviceEnergy16BRetrospectiveEnergyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_status_info import SmpmUlDeviceEnergy16BStatusInfoData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_tariff_consumed import \
    SmpmUlDeviceEnergy16BTariffConsumedData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_tariff_generated import \
    SmpmUlDeviceEnergy16BTariffGeneratedData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_gaz_flow_08b_daily import SmpmUlDeviceGazFlow08BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_gaz_flow_16b_daily import SmpmUlDeviceGazFlow16BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_gaz_flow_32b_daily import SmpmUlDeviceGazFlow32BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_heat_group_meter_56b_daily import SmpmUlDeviceHeatGroupMeter56BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_heat_group_meter_daily import SmpmUlDeviceHeatGroupMeterDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_heat_proxy_meter_16b_daily import SmpmUlDeviceHeatProxyMeter16BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_internal_info import SmpmUlDeviceInternalInfoData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter import SmpmUlDeviceJupiter08BCounterData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter_ch1 import SmpmUlDeviceJupiter08BCounterCh1Data
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter_ch2 import SmpmUlDeviceJupiter08BCounterCh2Data
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_08b_counter_volume import \
    SmpmUlDeviceJupiter08BCounterVolumeData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_12b_counter import SmpmUlDeviceJupiter12BCounterData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_12b_counter_volume import \
    SmpmUlDeviceJupiter12BCounterVolumeData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_jupiter_16b_counter import SmpmUlDeviceJupiter16BCounterData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_04b_event import SmpmUlDeviceWaterMeter04BEventData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_daily import SmpmUlDeviceWaterMeter08BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_daily_new import \
    SmpmUlDeviceWaterMeter08BDailyNewData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_08b_info import SmpmUlDeviceWaterMeter08BInfoData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_12b_daily import SmpmUlDeviceWaterMeter12BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_16b_daily import SmpmUlDeviceWaterMeter16BDailyData
from data_gateway_sdk.protocols.smpm.smpm_ul_device_water_meter_nbiot_16b_daily import \
    SmpmUlDeviceWaterMeterNbiot16BDailyData
from data_gateway_sdk.utils.packet import Packet


class SmpMGasMeterV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceGazFlow08BDailyData,
        SmpmUlDeviceGazFlow16BDailyData,
        SmpmUlDeviceGazFlow32BDailyData,
    )


class SmpMEnergyMeterV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceEnergy16B3PhaseConsumedData,
        SmpmUlDeviceEnergy16B3PhaseGeneratedData,
        SmpmUlDeviceEnergy16BDailyData,
        SmpmUlDeviceEnergy16BProfile8H1EnergyData,
        SmpmUlDeviceEnergy16BProfile8H2EnergyData,
        SmpmUlDeviceEnergy16BProfile8H3EnergyData,
        SmpmUlDeviceEnergy16BProfile8HArData,
        SmpmUlDeviceEnergy16BProfile8HFData,
        SmpmUlDeviceEnergy16BProfile8HPqsData,
        SmpmUlDeviceEnergy16BTariffConsumedData,
        SmpmUlDeviceEnergy16BTariffGeneratedData,
        SmpmUlDeviceEnergy16BInfoData,
        SmpmUlDeviceEnergy16BJournalData,
    )


class SmpMEnergyMeterV1DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceEnergy16B3PhaseConsumedData,
        SmpmUlDeviceEnergy16B3PhaseGeneratedData,
        SmpmUlDeviceEnergy16BDailyData,
        SmpmUlDeviceEnergy16BProfile8H1EnergyData,
        SmpmUlDeviceEnergy16BProfile8H2EnergyData,
        SmpmUlDeviceEnergy16BProfile8H3EnergyData,
        SmpmUlDeviceEnergy16BProfile8HArData,
        SmpmUlDeviceEnergy16BProfile8HFData,
        SmpmUlDeviceEnergy16BProfile8HPqsData,
        SmpmUlDeviceEnergy16BTariffConsumedData,
        SmpmUlDeviceEnergy16BTariffGeneratedData,
        SmpmUlDeviceEnergy16BInfoData,
        SmpmUlDeviceEnergy16BJournalData,
        SmpmUlDeviceEnergy16BElectricityParamsData,
        SmpmUlDeviceEnergy16BProfile8HEnergyData,
        SmpmUlDeviceEnergy16BRetrospectiveEnergyData,
        SmpmUlDeviceEnergy16BDeviceInfoData,
        SmpmUlDeviceEnergy16BStatusInfoData,
        SmpmUlDeviceDlAnswerData,
    )


class SmpMJupiter08BV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceJupiter08BCounterData,
        SmpmUlDeviceJupiter08BCounterCh1Data,
        SmpmUlDeviceJupiter08BCounterCh2Data,
        SmpmUlDeviceJupiter08BCounterVolumeData,
    )


class SmpMJupiter12BV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceJupiter12BCounterData,
        SmpmUlDeviceJupiter12BCounterVolumeData,
    )


class SmpMJupiter16BV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceJupiter16BCounterData,
    )


class SmpMWaterMeter04BV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceWaterMeter04BEventData,
    )


class SmpMWaterMeter08BV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceWaterMeter08BDailyData,
        SmpmUlDeviceWaterMeter08BDailyNewData,
        SmpmUlDeviceWaterMeter08BInfoData,
    )


class SmpMWaterMeter12BV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceWaterMeter12BDailyData,
    )


class SmpMWaterMeter16BV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceWaterMeter16BDailyData,
        SmpmUlDeviceWaterMeterNbiot16BDailyData,
    )


class SmpMHeatProxyMeter16BV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceHeatProxyMeter16BDailyData,
    )


class SmpMHeatGroupMeterV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceHeatGroupMeterDailyData,
        SmpmUlDeviceHeatGroupMeter56BDailyData,
    )


class SmpMInternalInfoDataV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmUlDeviceInternalInfoData,
    )


class SmpMDownlinkV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        SmpmDlDeviceEnergy8BGetDataData,
        SmpmDlDeviceEnergy8BSetClockData,
        SmpmDlDeviceEnergy8BSetRelayData,
        SmpmDlDeviceEnergy16BGetDataData,
        SmpmDlDeviceEnergy16BSetRegularDataSendingData,
    )



from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_8b_set_relay import SmpmDlDeviceEnergy8BSetRelayData, SmpmDlDeviceEnergy8bSetRelayId
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_dl_device_energy_8b_set_relay() -> None:
    case_serialized = bytes.fromhex("aa01000000000000")
    assert SmpmDlDeviceEnergy8BSetRelayData(state=False, relay_id=SmpmDlDeviceEnergy8bSetRelayId.ALL) == SmpmDlDeviceEnergy8BSetRelayData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("aa71000000000000")  # overflow
    assert SmpmDlDeviceEnergy8BSetRelayData(state=False, relay_id=SmpmDlDeviceEnergy8bSetRelayId.RELAY_ID_7) == SmpmDlDeviceEnergy8BSetRelayData.parse(BufRef(case_serialized))  # noqa: E501
